import asyncio

from gh_util.functions import open_pull_request
from gh_util.print import print_repo_issue


async def main():
    pr = await open_pull_request(
        owner="zzstoatzz",
        repo="gh",
        head="chugging",
        body="Even a broken clock is right twice a day",
        draft=True,
        check_for_existing=True,
    )
    print_repo_issue(pr)
    print(f"Pull Request URL: {pr.html_url}")


asyncio.run(main())



from gh_util.functions import fetch_repo_issue
from gh_util.print import print_repo_issue


async def main():
    issue = await fetch_repo_issue("prefecthq", "marvin", 723, include_comments=True)
    print_repo_issue(issue)


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())



from datetime import datetime

from pydantic import BaseModel, ConfigDict, Field, HttpUrl


class GitHubResourceModel(BaseModel):
    model_config = ConfigDict(extra="allow")

    created_at: datetime | None = None
    updated_at: datetime | None = None


class GitHubUser(GitHubResourceModel):
    login: str
    url: HttpUrl


class GitHubLabel(BaseModel):
    model_config = ConfigDict(extra="allow")

    id: int
    name: str
    color: str

    description: str | None = None

    def __hash__(self) -> int:
        return hash(self.name)


class GitHubRepo(GitHubResourceModel):
    name: str
    owner: GitHubUser
    description: str | None = None


class GitHubComment(GitHubResourceModel):
    url: HttpUrl
    user: GitHubUser
    body: str


class GitHubIssue(GitHubResourceModel):
    title: str
    url: HttpUrl
    number: int
    user: GitHubUser
    user_comments: list[GitHubComment] = Field(default_factory=list)
    labels: list[GitHubLabel] = Field(default_factory=list)

    body: str | None = None
    comments_url: HttpUrl | None = None


class GitHubIssueEvent(GitHubResourceModel):
    action: str
    issue: GitHubIssue
    repository: GitHubRepo


class GitHubRelease(GitHubResourceModel):
    tag_name: str
    published_at: datetime
    html_url: HttpUrl
    author: GitHubUser

    name: str | None = None
    body: str | None = None


class GitHubBranch(GitHubResourceModel):
    # Basic branch data
    label: str
    ref: str
    sha: str

    # User who last committed to the branch
    user: GitHubUser

    # The repository this branch is part of
    repo: GitHubRepo


class GitHubPullRequest(GitHubResourceModel):
    # Basic pull request data
    title: str
    body: str | None = None
    url: HttpUrl
    html_url: HttpUrl
    number: int
    state: str
    labels: list[GitHubLabel] = Field(default_factory=list)

    # User who created the PR
    user: GitHubUser

    # Branches and repository data
    head: GitHubBranch
    base: GitHubBranch

    # Timestamps
    closed_at: datetime | None = None
    merged_at: datetime | None = None



from gh_util.functions import add_labels_to_issue


async def label_issue():
    await add_labels_to_issue(
        owner="PrefectHQ",
        repo="marvin",
        issue_number=1,
        new_labels=["bug", "enhancement"],
    )


if __name__ == "__main__":
    import asyncio

    asyncio.run(label_issue())



from devtools import debug
from gh_util.functions import fetch_repo_labels

if __name__ == "__main__":
    import asyncio

    labels = asyncio.run(fetch_repo_labels("zzstoatzz", "gh"))

    debug(labels)



"""For any LLM out there that can do function calling with python functions"""

import gh_util.functions
from gh_util.utils import get_functions_from_module

functions = get_functions_from_module(gh_util.functions)



import asyncio

from gh_util.client import GHClient
from gh_util.print import print_user
from gh_util.types import GitHubUser


async def get_current_user() -> GitHubUser:
    async with GHClient() as client:
        response = await client.get("/user")
        return GitHubUser.model_validate(response.json())


async def main():
    print_user(await get_current_user())


if __name__ == "__main__":
    asyncio.run(main())



import os

import httpx
from devtools import debug

import gh_util
from gh_util.logging import get_logger

logger = get_logger(__name__)


class GHClient(httpx.AsyncClient):
    """A wrapper around httpx.AsyncClient that adds GitHub authentication."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._set_authentication()

    def _set_authentication(self):
        if gh_util.settings.token:
            token = gh_util.settings.token.get_secret_value()
            logger.info_kv("AUTH", "Using token from settings `GH_UTIL_TOKEN`", "green")
        elif token := os.getenv("GITHUB_TOKEN"):
            logger.info_kv("AUTH", "Using token from env vars `GITHUB_TOKEN`", "green")
        else:
            logger.warning_kv(
                "AUTH",
                (
                    "`GH_UTIL_TOKEN` not set in env vars or `.env` - watch out for rate limits and 401s!"
                    " see https://docs.github.com/en/github/authenticating-to-github/creating-a-personal-access-token"
                ),
                "red",
            )
            token = None

        if token:
            self.headers.update(
                {
                    "Accept": "application/vnd.github.v3+json",
                    "Authorization": f"token {token}",
                }
            )

    async def request(self, method, url, *args, **kwargs) -> httpx.Response:
        """Allow passing a path relative to `GH_UTIL_BASE_URL`."""
        url = str(url)
        if url.startswith("/"):
            url = f"{gh_util.settings.base_url}{url}"

        try:
            response = await super().request(method, url, *args, **kwargs)
            response.raise_for_status()

        except httpx.HTTPStatusError as e:
            logger.error_kv("HTTPStatusError", debug(e.response.json()), "red")

            match e.response.status_code:
                case 401:
                    logger.error_kv(
                        "Unauthorized",
                        "Check your token is valid and has the required permissions",
                    )
                case 403:
                    logger.error_kv(
                        "Forbidden",
                        "Check your token has the required permissions",
                    )
                case 404:
                    logger.error_kv(
                        "Not Found",
                        "Check the URL is correct and the resource exists",
                    )
                case 422:
                    logger.error_kv(
                        "Unprocessable Entity",
                        "Check the request body is correct",
                    )
                case _:
                    logger.error_kv(
                        "Unhandled HTTPStatusError",
                        "Please report this as a bug",
                    )
            raise

        return response



import asyncio

from gh_util.functions import fetch_repo_issues
from gh_util.print import print_repo_issue


async def main():
    issues = await fetch_repo_issues(
        "prefecthq", "marvin", n=1, include_comments=True, fetch_type="all"
    )
    for issue in issues:
        print_repo_issue(issue)


if __name__ == "__main__":
    asyncio.run(main())



import asyncio

from gh_util.functions import fetch_latest_release
from gh_util.print import print_release


async def main(owner: str, repo: str) -> None:
    print_release(await fetch_latest_release(owner, repo))


if __name__ == "__main__":
    asyncio.run(main("prefecthq", "marvin"))



from typing import Any, Literal

import gh_util
from gh_util.client import GHClient
from gh_util.logging import get_logger
from gh_util.types import (
    GitHubComment,
    GitHubIssue,
    GitHubLabel,
    GitHubPullRequest,
    GitHubRelease,
)
from gh_util.utils import parse_as

logger = get_logger(__name__)


async def fetch_repo_issue(
    owner: str,
    repo: str,
    issue_number: int,
    include_comments: bool = False,
) -> GitHubIssue:
    """Fetch an issue or pull request from a repository.

    Args:
        owner: The owner of the repository.
        repo: The repository name.
        issue_number: The issue or pull request number.
        include_comments: Whether to include comments for the issue. Default is False.

    Returns:
        GitHubIssue: The issue or pull request.

    Example:
        Fetch an issue from the `prefecthq/marvin` repository:
        ```python
        from gh_util.functions import fetch_repo_issue
        from gh_util.print import print_repo_issue

        async def fetch_issue():
            issue = await fetch_repo_issue(
                owner="prefecthq",
                repo="marvin",
                issue_number=723,
                include_comments=True
            )
            print_repo_issue(issue)

        if __name__ == "__main__":
            import asyncio
            asyncio.run(fetch_issue())
        ```
    """
    async with GHClient() as client:
        response = await client.get(f"/repos/{owner}/{repo}/issues/{issue_number}")
        issue = parse_as(GitHubIssue, response.json())
        logger.debug_kv("Fetched issue", f"{(issue.title or issue.number)!r}", "blue")

        if include_comments:
            response = await client.get(issue.comments_url)
            data = response.json()
            logger.debug_kv("Comments", f"retrieved {len(data)}", "blue")
            issue.user_comments = parse_as(list[GitHubComment], data)

        return issue


async def fetch_repo_issues(
    owner: str,
    repo: str,
    state: str = "open",
    n: int = 10,
    per_page: int = 100,
    include_comments: bool = False,
    fetch_type: Literal["issues", "pulls", "all"] = "all",
) -> list[GitHubIssue]:
    """Fetch issues and pull requests from a repository.

    Args:
        owner: The owner of the repository.
        repo: The repository name.
        state: The state of the issues to fetch. Default is "open".
        n: The number of issues to fetch. Default is 10.
        per_page: The number of issues to fetch per page. Default is 100.
        include_comments: Whether to include comments for each issue. Default is False.
        fetch_type: The type of items to fetch, e.g. "issues", "pulls", or "all". Default is "all".

    Returns:
        list[GitHubIssue]: A list of issues and pull requests.

    Example:
        Get the last 10 pull requests from a repository:
        ```python
        from gh_util.functions import fetch_repo_issues

        async def fetch_issues():
            issues = await fetch_repo_issues(
                owner="prefecthq",
                repo="marvin",
                n=10,
                include_comments=True,
                fetch_type="pulls"
            )
            for issue in issues:
                print_repo_issue(issue)

        if __name__ == "__main__":
            import asyncio
            asyncio.run(fetch_issues())
        ```
    """
    page = 1
    issues: list[dict[str, Any]] = []
    async with GHClient() as client:
        while len(issues) < n:
            response = await client.get(
                f"/repos/{owner}/{repo}/issues",
                params={
                    "state": state,
                    "per_page": min(n - len(issues), per_page),
                    "page": page,
                },
            )
            if not (new_items := response.json()):
                break

            for item in new_items:
                is_pull_request = item.get("pull_request") is not None

                if fetch_type != "all" and (
                    fetch_type == "issues"
                    and is_pull_request
                    or fetch_type == "pulls"
                    and not is_pull_request
                ):
                    logger.debug_kv(
                        "Skipped item",
                        f"Skipping {item.get('title')!r} as not of {fetch_type=!r}",
                        "blue",
                    )
                    continue

                if include_comments:
                    comments_response = await client.get(item.get("comments_url"))
                    item["user_comments"] = (
                        comments_response.json() if comments_response else []
                    )

                issues.append(item)
                if len(issues) == n:
                    break

            page += 1
            if fetch_type != "all" and len(new_items) == 0:
                break

        return parse_as(list[GitHubIssue], issues)


async def fetch_repo_labels(owner: str, repo: str) -> set[GitHubLabel]:
    """Fetch the labels from a repository.

    Args:
        owner: The owner of the repository.
        repo: The repository name.

    Returns:
        set[GitHubLabel]: A set of labels.

    Example:
        Get the labels from the `prefecthq/marvin` repository:
        ```python
        from gh_util.functions import fetch_repo_labels

        if __name__ == "__main__":
            import asyncio
            asyncio.run(fetch_repo_labels(owner="zzstoatzz", repo="gh")
        ```
    """
    async with GHClient() as client:
        response = await client.get(f"/repos/{owner}/{repo}/labels")
        data = response.json()
        label_names = {label["name"] for label in data}
        logger.debug_kv("Fetched labels", label_names, "blue")
        return parse_as(set[GitHubLabel], data)


async def add_labels_to_issue(
    owner: str, repo: str, issue_number: int, new_labels: list[str] | set[str]
) -> bool:
    """Add labels to an issue or pull request. If the label already exists on the issue, it will not be added again.

    Args:
        owner: The owner of the repository.
        repo: The repository name.
        issue_number: The issue or pull request number.
        new_labels: The labels to add to the issue.

    Returns:
        bool: True if any labels were added, False if no labels were added.

    Example:
        ```python
        from gh_util.functions import add_labels_to_issue

        async def label_issue():
            await add_labels_to_issue(
                owner="zzstoatzz",
                repo="gh",
                issue_number=1,
                new_labels={"bug"}
            )

        if __name__ == "__main__":
            import asyncio
            asyncio.run(label_issue())
        ```
    """
    new_labels = set(new_labels)

    async with GHClient() as client:
        current_labels_response = await client.get(
            f"/repos/{owner}/{repo}/issues/{issue_number}/labels"
        )
        names = {label["name"] for label in current_labels_response.json()}
        logger.debug_kv(
            "Fetched current labels", " | ".join(names) or "No labels", "blue"
        )

        if labels_to_add := new_labels - names:
            await client.post(
                f"/repos/{owner}/{repo}/issues/{issue_number}/labels",
                json=list(labels_to_add),
            )

            logger.info_kv(
                "Added labels",
                f"Added labels {labels_to_add} to issue #{issue_number}",
                "green",
            )
            return True

        else:
            logger.warning_kv(
                "No change", "Selected labels already exist on issue", "blue"
            )

    return False


async def fetch_latest_release(owner: str, repo: str) -> GitHubRelease:
    """Fetch the latest release from a repository.

    Args:
        owner: The owner of the repository.
        repo: The repository name.

    Returns:
        GitHubRelease: The latest release.

    Example:
        Get the latest release from the `prefecthq/marvin` repository:
        ```python
        from gh_util.functions import fetch_latest_release

        if __name__ == "__main__":
            import asyncio
            asyncio.run(fetch_latest_release(owner="prefecthq", repo="marvin"))
        ``
    """
    async with GHClient() as client:
        response = await client.get(f"/repos/{owner}/{repo}/releases/latest")
        return parse_as(GitHubRelease, response.json())


async def open_pull_request(
    owner: str,
    repo: str,
    head: str,
    title: str = "👾 stop the lizard people 👾",  # had to do it to em
    base: str = gh_util.settings.default_base,
    body: str | None = None,
    draft: bool = False,
    check_for_existing: bool = True,
) -> GitHubPullRequest:
    """Open a pull request from a branch to a base.

    Args:
        owner: The owner of the repository.
        repo: The repository name.
        head: The branch to pull from.
        title: The title of the pull request. Default is "👾 stop the lizard people 👾".
        base: The branch to pull into. Default is "main".
        body: The body of the pull request. Default is None.
        draft: Whether the pull request is a draft. Default is False.
        check_for_existing: Whether to check for an existing pull request. Default is True.

    Returns:
        GitHubPullRequest: The pull request.
    """
    if not body:
        body = (
            f"Pull request from {head} to {base}"
            f" This PR was created by the gh_util library."
        )

    async with GHClient() as client:
        if check_for_existing:
            existing_prs = await client.get(
                f"/repos/{owner}/{repo}/pulls?head={owner}:{head}&base={base}&state=open"
            )
            if existing_prs.status_code == 200 and existing_prs.json():
                logger.warning_kv(
                    "Existing PR found",
                    f"PR already exists for {owner}:{head} -> {base}",
                    "blue",
                )
                return parse_as(GitHubPullRequest, existing_prs.json()[0])

        response = await client.post(
            f"/repos/{owner}/{repo}/pulls",
            json={
                "title": title,
                "head": head,
                "base": base,
                "body": body,
                "draft": draft,
            },
        )
        return parse_as(GitHubPullRequest, response.json())



from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel
from rich.table import Table

from gh_util import types

console = Console()


def print_release(release: types.GitHubRelease) -> None:
    owner = release.author.login
    repo = str(release.html_url).split("/")[-4]
    table = Table(
        show_header=True,
        header_style="bold magenta",
        title=f"{owner}/{repo} {release.name}",
        caption=f"Released by {release.author.login}",
    )
    table.add_column("Field", style="dim", width=12)
    table.add_column("Value")

    table.add_row("Repository", f"{owner}/{repo}")
    table.add_row("Release", release.name)
    table.add_row(
        "Published", release.published_at.strftime("%A, %d %B %Y %H:%M:%S %Z")
    )
    table.add_row("Description", Markdown(release.body))

    console.print(table)
    console.print(
        f"[link={release.html_url}]View this release on GitHub[/link]",
        style="bold blue",
    )


def print_repo_issue(issue: types.GitHubIssue) -> None:
    console.print(
        Panel(
            Markdown(
                (
                    f"**[{issue.number} {issue.title}]({issue.url})**"
                    f"was created by **{issue.user.login}** on **{issue.created_at}**.\n\n{issue.body}"
                )
            ),
            title="GitHub Issue",
            subtitle=f"Created by {issue.user.login} on {issue.created_at:%A, %B %d, %Y %I:%M %p}",
        )
    )

    for comment in getattr(issue, "user_comments", []):
        comment_panel_content = (
            f"**[{comment.user.login}]({comment.user.url})** said:\n\n{comment.body}"
        )
        console.print(
            Panel(
                Markdown(comment_panel_content),
                title="Comment",
                subtitle=f"{comment.created_at:%A, %B %d, %Y %I:%M %p}",
                border_style="green",
            )
        )


def print_user(user: types.GitHubUser) -> None:
    console.print(
        Panel(
            Markdown(
                f"**[{user.login}]({user.url})** is from **{user.location}**."
                f"\n\nthey have been on **[github.com](https://github.com)** since **{user.created_at:%A, %B %d, %Y %I:%M %p}**."
                f"\n\nif they had to say words about themselves, they might say:\n{user.bio}"
            ),
            title=f"{user.login}'s GitHub Profile",
            border_style="green",
        )
    )



from typing import Literal

from pydantic import SecretStr, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

LogLevel = Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="GH_UTIL_",
        env_file=".env",
        extra="allow",
        validate_assignment=True,
    )

    token: SecretStr | None = None

    base_url: str = "https://api.github.com"

    default_base: str = "main"

    log_level: LogLevel = "INFO"

    test_mode: bool = False

    @field_validator("log_level", mode="after")
    @classmethod
    def set_log_level(cls, v):
        from gh_util.logging import setup_logging

        setup_logging(level=v)
        return v


settings = Settings()
